indexing

   description: "Class that covers a path on the file system.%
   %Also makes possible to use either a / or \ as directory separator%
   %(Unix or Windows convention).%
   %Optionally, it can do environment variable expansion for names as%
   %$(TMP), ${TMP}, $TMP or %%TMP%% (read single percent sign here!).%
   %Also recognizes ~ and replaces by $HOME except if ~ is followed by a%
   %digit, so it should work more or less on Windows when it should, and%
   %do nothing when it shouldn't."

   author: "Berend de Boer"
   date: "$Date: 2001/12/27 $";
   revision: "$Revision: #1 $"

class

   STDC_PATH


inherit

   STRING
      rename
         make_from_string as org_make_from_string
      end


creation

   make,
   make_from_string,
   make_expand


feature

   make_from_string (a_path: STRING) is
         -- the new path contains the correct directory separator
         -- independent of what is used in `s'
      do
         org_make_from_string (a_path)
         correct_directory_separator
         -- support Windooze, which doesn't like a trailing slash
         remove_trailing_slash
      end

   make_expand (s: STRING) is
         -- the new path is the expanded contents of `s'
      do
         org_make_from_string (normalize_path (s))
         -- support Windooze, which doesn't like a trailing slash
         remove_trailing_slash
      end


feature -- portability queries

   is_portable_character (c: CHARACTER): BOOLEAN is
         -- return True if c is a POSIX portable character, i.e. it
         -- can safely be used in filenames
      do
         Result :=
            (c >= 'a' and c <= 'z') or else
            (c = '.') or else
            (c = '_') or else
            (c >= '0' and c <= '9') or else
            (c >= 'A' and c <= 'Z') or else
            (c = '-');
      end

   is_portable: BOOLEAN is
         -- Return True if this path is portable
      local
         i: INTEGER
      do
         from
            i := 1
            Result := True
         until
            not Result or i > count
         loop
            Result := is_portable_character (item (i))
               or else item (i) = directory_separator
            i := i + 1
         end
      end


feature -- compare to string

   is_equal_to_string (s: STRING): BOOLEAN is
      do
         Result := s.count = count and then s.substring_index (Current, 1) = 1
      end


feature -- path individual components

   directory,
   basename,
   suffix: STRING
         -- individual components of path, only valid after `parse'
         -- has been called. Not automagically updated

   parse (suffix_list: ARRAY[STRING]) is
         -- set `directory', `basename', and `suffix'
         -- the suffix should be in `suffix_list' if it needs to be stripped
      local
         slash_position,
         suffix_position: INTEGER
         i: INTEGER
         found: BOOLEAN
      do
         -- dir
         from
            slash_position := count
            found := False
         until
            found or slash_position = 0
         loop
            found := item (slash_position) = directory_separator
            slash_position := slash_position - 1
         end
         if  found then
            slash_position := slash_position + 1
            create directory.make_from_string (substring (1, slash_position - 1))
         else
            slash_position := 0
            directory := ""
         end

         -- suffix
         found := suffix_list /= Void
         if found then
            from
               i := suffix_list.lower
               found := False
            until
               found or i > suffix_list.upper
            loop
               suffix := suffix_list.item (i)
               suffix_position := count - suffix.count + 1
               found := substring_index (suffix, suffix_position) = suffix_position
               if not found then
                  i := i + 1
               end
            end
         end
         if found then
            suffix := clone (suffix)
         else
            suffix_position := count + 1
            suffix := ""
         end

         -- basename
         create basename.make_from_string (substring (slash_position + 1, suffix_position - 1))
      ensure
         directory_set: directory /= Void
         suffix_set: suffix /= Void
         only_suffix_when_needed: suffix_list = Void implies suffix.is_empty
      end

   set_directory (new_directory: STRING) is
         -- build new path from `new_directory', `basename' and `suffix'
      do
         wipe_out
         if new_directory /= Void and then not new_directory.is_empty then
            append_string (new_directory)
            if new_directory.item (new_directory.count) /= directory_separator then
               append_character (directory_separator)
            end
         end
         append_string (basename)
         append_string (suffix)
         if new_directory = Void then
            create directory.make (0)
         else
            directory := clone (new_directory)
         end
      ensure
         directory_set: new_directory /= Void implies directory.is_equal (new_directory)
      end

   set_suffix (new_suffix: STRING) is
         -- build new path from current `directory', `basename' and `new_suffix'
      do
         wipe_out
         if directory /= Void and then not directory.is_empty then
            append_string (directory)
            append_character (directory_separator)
         end
         append_string (basename)
         append_string (new_suffix)
         suffix := clone (new_suffix)
      ensure
         suffix_set: suffix.is_equal (new_suffix)
      end


feature -- when path is a directory

   remove_trailing_slash is
      local
         has_trailing_slash: BOOLEAN
      do
         has_trailing_slash := count > 0 and item (count) = directory_separator
         if has_trailing_slash then
            remove (count)
         end
      ensure
         no_trailing_slash: count = 0 or else
                            item (count) /= directory_separator
      end


feature {NONE}

   correct_directory_separator is
         -- turn a '/' into a '\' or vice versa depending on the
         -- current platform
      local
         i: INTEGER
         c: CHARACTER
      do
         from
            i := 1
         until
            i > count
         loop
            c := item (i)
            inspect c
            when '/', '\' then
               put (directory_separator, i)
            else -- ok
            end
            i := i + 1
         end
      end

   normalize_path (s: STRING): STRING is
      local
         i, j: INTEGER
         start_name, stop_name: INTEGER
         c: CHARACTER
         env_name: STRING
      do
         -- Result := sh.make_with_capacity (s.count * 2)
         create Result.make (s.count * 2)
         Result.wipe_out
         from
            i := 1
         until
            i > s.count
         loop
            c := s.item (i)
            inspect c
            when '~' then
               if (i = 1) and then (s.count > 1) and then
                  not is_digit (s.item (i+1)) then
                  Result.append_string (home_directory)
               else
                  Result.append_character (c)
               end
            when '\' then
               Result.append_character (directory_separator)
            when '/' then
               Result.append_character (directory_separator)
            when '$' then
               if (i < s.count) then
                  inspect s.item (i+1)
                  when '{' then
                     start_name := i+2
                     j := s.index_of ('}', start_name)
                     stop_name := j-1
                     i := j
                  when '(' then
                     start_name := i+2
                     j := s.index_of (')', start_name)
                     stop_name := j-1
                     i := j
                  when '$' then
                     -- on Windows you can have temp names like tmp1.$$$
                     Result.append_character (c)
                  else
                     start_name := i+1
                     from
                        j := start_name
                     until
                        j > s.count or else not is_env_character (s.item(j))
                     loop
                        j := j + 1
                     end
                     stop_name := j-1
                     i := j-1
                  end
                  if start_name <= stop_name - 1 then
                     env_name := s.substring (start_name, stop_name)
                     Result.append_string (env_value (env_name))
                  end
               else
                  -- if last character is a $, do not consider it a
                  -- special character
                  Result.append_character (c)
               end
            when '%%' then
               start_name := i+1
               j := s.index_of ('%%', start_name)
               stop_name := j-1
               i := j
               env_name := s.substring (start_name, stop_name)
               Result.append_string (env_value (env_name))
            else
               Result.append_character (c)
            end
            i := i + 1
         end
      end

   directory_separator: CHARACTER is
         -- attempts to return a platform specific directory separator
      once
         if is_windows then
            Result := '\'
         else
            Result := '/'
         end
      end

   env_value (env_name: STRING): STRING is
         -- return value of environment variable `env_name'
      local
         e: STDC_ENV_VAR
      do
         create e.make (env_name)
         Result := e.value
      end

   home_directory: STRING is
      local
         env_home: STDC_ENV_VAR
      do
         create env_home.make ("HOME")
         Result := env_home.value
      end

   is_digit (c: CHARACTER): BOOLEAN is
         -- return True if c is a digit, it seems with Eiffel you can
         -- write every shitty routine yourself it you want compatibility
      do
         Result := (c >= '0') and (c <= '9');
      end

   is_env_character (c: CHARACTER): BOOLEAN is
         -- return True if c can be used in an environment variable
         -- used by our expansion routine.
      do
         Result :=
            (c >= 'A' and c <= 'Z') or else
            (c >= 'a' and c <= 'z') or else
            (c >= '0' and c <= '9') or else
            (c = '_');
      end


feature {NONE}

   is_windows: BOOLEAN is
      external "C"
      end


end -- class STDC_PATH
